@@ -165,22 +165,56 @@ ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
     return YKPIV_PCSC_ERROR;
   }
 
-  reader_ptr = reader_buf;
-  if(wanted) {
-    while(*reader_ptr != '\0') {
-      if(strstr(reader_ptr, wanted)) {
-	if(state->verbose) {
-	  fprintf(stderr, "using reader '%s' matching '%s'.\n", reader_ptr, wanted);
-	}
-	break;
+  for(reader_ptr = reader_buf; *reader_ptr != '\0'; reader_ptr += strlen(reader_ptr) + 1) {
+    if(wanted) {
+      if(!strstr(reader_ptr, wanted)) {
+        if(state->verbose) {
+          fprintf(stderr, "skipping reader '%s' since it doesn't match '%s'.\n", reader_ptr, wanted);
+        }
+        continue;
+      }
+    }
+    if(state->verbose) {
+      fprintf(stderr, "trying to connect to reader '%s'.\n", reader_ptr);
+    }
+    rc = SCardConnect(state->context, reader_ptr, SCARD_SHARE_SHARED,
+        SCARD_PROTOCOL_T1, &state->card, &active_protocol);
+    if(rc != SCARD_S_SUCCESS)
+    {
+      if(state->verbose) {
+        fprintf(stderr, "SCardConnect failed, rc=%08lx\n", rc);
+      }
+      continue;
+    }
+
+    {
+      APDU apdu;
+      unsigned char data[0xff];
+      unsigned long recv_len = sizeof(data);
+      int sw;
+      ykpiv_rc res;
+
+      memset(apdu.raw, 0, sizeof(apdu));
+      apdu.st.ins = 0xa4;
+      apdu.st.p1 = 0x04;
+      apdu.st.lc = sizeof(aid);
+      memcpy(apdu.st.data, aid, sizeof(aid));
+
+      if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
+        if(state->verbose) {
+          fprintf(stderr, "Failed communicating with card: '%s'\n", ykpiv_strerror(res));
+        }
+        continue;
+      } else if(sw == 0x9000) {
+        return YKPIV_OK;
       } else {
-	if(state->verbose) {
-	  fprintf(stderr, "skipping reader '%s' since it doesn't match.\n", reader_ptr);
-	}
-	reader_ptr += strlen(reader_ptr) + 1;
+        if(state->verbose) {
+          fprintf(stderr, "Failed selecting applet: %04x\n", sw);
+        }
       }
     }
   }
+
   if(*reader_ptr == '\0') {
     if(state->verbose) {
       fprintf(stderr, "error: no useable reader found.\n");
@@ -189,40 +223,7 @@ ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
     return YKPIV_PCSC_ERROR;
   }
 
-  rc = SCardConnect(state->context, reader_ptr, SCARD_SHARE_SHARED,
-      SCARD_PROTOCOL_T1, &state->card, &active_protocol);
-  if(rc != SCARD_S_SUCCESS)
-  {
-    if(state->verbose) {
-      fprintf(stderr, "error: SCardConnect failed, rc=%08lx\n", rc);
-    }
-    SCardReleaseContext(state->context);
-    return YKPIV_PCSC_ERROR;
-  }
-
-  {
-    APDU apdu;
-    unsigned char data[0xff];
-    unsigned long recv_len = sizeof(data);
-    int sw;
-    ykpiv_rc res;
-
-    memset(apdu.raw, 0, sizeof(apdu));
-    apdu.st.ins = 0xa4;
-    apdu.st.p1 = 0x04;
-    apdu.st.lc = sizeof(aid);
-    memcpy(apdu.st.data, aid, sizeof(aid));
-
-    if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
-      return res;
-    } else if(sw == 0x9000) {
-      return YKPIV_OK;
-    }
-
-    return YKPIV_APPLET_ERROR;
-  }
-
-  return YKPIV_OK;
+  return YKPIV_GENERIC_ERROR;
 }
 
 ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
@@ -243,8 +244,8 @@ ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
   }
   do {
     size_t this_size = 0xff;
-    unsigned long recv_len = 0xff;
-    unsigned char data[0xff];
+    unsigned char data[261];
+    unsigned long recv_len = sizeof(data);
     APDU apdu;
 
     memset(apdu.raw, 0, sizeof(apdu.raw));
@@ -280,8 +281,8 @@ ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
   } while(in_ptr < in_data + in_len);
   while(*sw >> 8 == 0x61) {
     APDU apdu;
-    unsigned long recv_len = 0xff;
-    unsigned char data[0xff];
+    unsigned char data[261];
+    unsigned long recv_len = sizeof(data);
 
     if(state->verbose > 2) {
       fprintf(stderr, "The card indicates there is %d bytes more data for us.\n", *sw & 0xff);
@@ -347,7 +348,7 @@ static ykpiv_rc send_data(ykpiv_state *state, APDU *apdu,
 
 ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
   APDU apdu;
-  unsigned char data[0xff];
+  unsigned char data[261];
   DES_cblock challenge;
   unsigned long recv_len = sizeof(data);
   int sw;
@@ -390,7 +391,7 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
     DES_cblock response;
     DES_ecb3_encrypt(&challenge, &response, &ks1, &ks2, &ks3, 0);
 
-    recv_len = 0xff;
+    recv_len = sizeof(data);
     memset(apdu.raw, 0, sizeof(apdu));
     apdu.st.ins = YKPIV_INS_AUTHENTICATE;
     apdu.st.p1 = YKPIV_ALGO_3DES; /* triple des */
@@ -433,7 +434,7 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
 
 ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
   APDU apdu;
-  unsigned char data[0xff];
+  unsigned char data[261];
   unsigned long recv_len = sizeof(data);
   int sw;
   size_t i;
@@ -628,7 +629,7 @@ ykpiv_rc ykpiv_decipher_data(ykpiv_state *state, const unsigned char *in,
 
 ykpiv_rc ykpiv_get_version(ykpiv_state *state, char *version, size_t len) {
   APDU apdu;
-  unsigned char data[0xff];
+  unsigned char data[261];
   unsigned long recv_len = sizeof(data);
   int sw;
   ykpiv_rc res;
@@ -650,7 +651,7 @@ ykpiv_rc ykpiv_get_version(ykpiv_state *state, char *version, size_t len) {
 
 ykpiv_rc ykpiv_verify(ykpiv_state *state, const char *pin, int *tries) {
   APDU apdu;
-  unsigned char data[0xff];
+  unsigned char data[261];
   unsigned long recv_len = sizeof(data);
   int sw;
   size_t len = 0;
@@ -679,7 +680,7 @@ ykpiv_rc ykpiv_verify(ykpiv_state *state, const char *pin, int *tries) {
   } else if(sw == 0x9000) {
     return YKPIV_OK;
   } else if((sw >> 8) == 0x63) {
-    *tries = (sw & 0xff);
+    *tries = (sw & 0xf);
     return YKPIV_WRONG_PIN;
   } else if(sw == 0x6983) {
     *tries = 0;
